#!/usr/bin/env python3
import matplotlib.pyplot as plt
import numpy as np
import pathlib

from sklearn.model_selection import train_test_split
from sklearn.cross_decomposition import PLSRegression

from OPLS_MD import PLS, OPLS, OPLS_PLS, PLS_MD


with np.load(pathlib.Path(__file__).parent.parent / "rsc" / "test_data.npz") as npz:
    X = npz["X"]
    y = npz["y"]
    Y = npz["Y"]


X_train, X_test, y_train, y_test = train_test_split(
    X, y, test_size=0.5, random_state=42)

if (True):
    import MDAnalysis as mda
    u = mda.Universe(
        str(pathlib.Path(__file__).parent.parent / "rsc" / "tmd.pdb")
    )
    u.load_new(X_train.reshape((X_train.shape[0], u.atoms.n_atoms, 3)))

    pls_md = PLS_MD(n_components=5).fit(u, y_train)
    pls = PLS(n_components=5).fit(X_train, y_train)
    pls_flip = PLS(n_components=5, flip=True).fit(X_train, y_train)
    pls_sklearn = PLSRegression(n_components=5).fit(X_train, y_train)
    u.load_new(X_test.reshape((X_test.shape[0], u.atoms.n_atoms, 3)))
    assert np.all(np.abs(pls_md.transform(u)-pls.transform(X_test)) < 1e-6)
    assert np.abs(pls_md.score(u, y_test)-pls.score(X_test, y_test)) < 1e-10
    assert np.all(
        np.abs(pls_sklearn.transform(X_test)-pls_flip.transform(X_test)) < 1e-6
    )
    assert np.abs(pls_sklearn.score(X_test, y_test) -
                  pls_flip.score(X_test, y_test)) < 1e-10
    assert np.abs(pls_sklearn.score(X_test, y_test) -
                  pls.score(X_test, y_test)) < 1e-10

quit()
if (True):
    fig1, axes1 = plt.subplots(1, 2)
    fig2, axes2 = plt.subplots(1, 2)

    for center, scale in ((True, True), (True, False), (False, False)):
        print(f"center={center}, scale={scale}")
        ncomp = []
        opls = []
        pls = []
        for i in range(1, 16):
            ncomp.append(i)
            opls.append(OPLS_PLS(n_components=1, pls_components=i, center=center,
                        scale=scale).fit(X_train, y_train).score(X_test, y_test))
            pls.append(PLS(n_components=i, center=center, scale=scale
                           ).fit(X_train, y_train).score(X_test, y_test))

        ncomp = np.array(ncomp)
        axes1[0].plot(ncomp, pls,  label=f"center={center}, scale={scale}")
        axes1[1].plot(ncomp, opls, label=f"center={center}, scale={scale}")

        axes2[0].plot(ncomp-((not center) and (not scale)), pls,
                      label=f"center={center}, scale={scale}")
        axes2[1].plot(ncomp-((not center) and (not scale)), opls,
                      label=f"center={center}, scale={scale}")

        for axes in (axes1, axes2):
            axes[0].set_title("Pure PLS")
            axes[1].set_title("OPLS filtering with one component")

            for ax in axes:
                ax.legend()
                ax.set_xlabel("number of components")
                ax.set_ylabel("Test score")

    for fig in (fig1, fig2):
        fig.set_size_inches(14, 8)
        fig.tight_layout()
    fig1.savefig(pathlib.Path(__file__).parent / "center_scale_test1.png")
    fig2.savefig(pathlib.Path(__file__).parent / "center_scale_test2.png")

    # quit()

if (True):
    maxk = 15

    ncomp = np.arange(maxk)+1
    pls_score = []
    for k in ncomp:
        print(k, end="\r")
        pls = PLS(n_components=k).fit(X_train, y_train)
        pls_score.append(pls.score(X_test, y_test))

    print("pls ", ["%.4f" % v for v in pls_score])

    ncomp = np.arange(maxk)+1
    pls_score = []
    pls = PLS(n_components=maxk).fit(X_train, y_train)
    for k in ncomp:
        print(k, end="\r")
        pls_score.append(pls.score(X_test, y_test, ndim=k))

    print("pls ", ["%.4f" % v for v in pls_score])

    opls_pls_score = {}
    ncomp_opls = np.arange(1, 7)
    for i in ncomp_opls:
        opls_pls_score[i] = []
        opls = OPLS(n_components=i).fit(X_train, y_train)
        pls = PLS(n_components=maxk).fit(opls.correct(X_train), y_train)
        for k in ncomp:
            print(i, k, end="\r")
            opls_pls_score[i].append(
                pls.score(opls.correct(X_test), y_test, ndim=k))

        print("%2d  " % i, ["%.4f" % v for v in opls_pls_score[i]])

    fig, ax = plt.subplots(1)
    ax.plot(ncomp, pls_score, label="PLS")
    for i in opls_pls_score:
        ax.plot(ncomp, opls_pls_score[i], "--", label=f"OPLS({i})-PLS")

    ax.legend()
    fig.set_size_inches(12, 8)
    fig.tight_layout()
    fig.savefig(pathlib.Path(__file__).parent / "scores.png")

    fig, ax = plt.subplots(1)
    ax.plot(ncomp, pls_score, label="PLS")
    for i in opls_pls_score:
        ax.plot(ncomp+i, opls_pls_score[i], "--", label=f"OPLS({i})-PLS")

    ax.set_xlim(ncomp.min()-1, ncomp.max()+1)
    ax.legend()
    fig.set_size_inches(12, 8)
    fig.tight_layout()
    fig.savefig(pathlib.Path(__file__).parent / "scores2.png")

jaas = 10
print(0, jaas,
      PLS(n_components=10).fit(X_train, y_train).score(X_test, y_test))
for i in range(1, jaas):
    k = jaas-i
    opls = OPLS(n_components=i).fit(X_train, y_train)
    pls = PLS(n_components=k).fit(opls.correct(X_train), y_train)
    print(i, k, pls.score(opls.correct(X_test), y_test),
          OPLS_PLS(i, k).fit(X_train, y_train).score(X_test, y_test))



from typing import Optional, overload
import numpy as np

from .PLS import PLS
from .OPLS import OPLS


class OPLS_PLS(OPLS):
    """
    Orthogonal Projection on Latent Structure (O-PLS) wrapper for a Partial Least Squares (PLS)  model.

    Parameters
    ----------
    n_components : int, default=1
        Number of components to fit.
    pls_components : int, default=2
        Number of components to fit.
    scale : bool, default=True
        Whether to scale X and Y to unit variance
    flip : bool, default=False
        Whether to flip the singular vectors for compatibility with different solvers.
        With flip=True and deflation_mode="regression" the first PLS component will be
        exactly (up to machine precision) the same as with sklearn.cross_decomposition.PLSRegression.
        Does not affect the results in any other meaningful way.
    max_iter : int, default=500
        Maximum number of NIPALS iterations.
    tol : float, default=1e-06
        Tolerance for stopping NIPALS iterations.
    copy : bool, default=True
        Whether to make copies of X and Y. False does not guarantee everything is in place, but
        True does guarantee copying.
    dtype : numpy.dtype, default=float64
        numpy dtype to cast input to. If None calculations will be done with whatever dtype the input arrays have.
    algorithm : str, default="OPLS"
        The algorithm to use. Acceptable values are "OPLS" and "O2PLS".
        NOTE: "O2PLS" is not yet well tested, and OPLS is only tested with univariate y.
    deflation_mode : str, default=None
        Whether to calculate y deflation with x_scores ("regression") or y_scores ("canonical").
        The latter only is reliable with n_components<=Y.shape[1].
        With OPLS algorithm this only changes the y-loadings, y-rotations and the final regressor, so
        it does not affect the corrected coordinates.
        With O2PLS algorithm "canonical" should be used.
        If None, "regression" is used for OPLS and "canonical" with O2PLS.

    Attributes
    ----------
    predictive_scores: np.ndarray
        Predictive x-scores.
    predictive_loadings: np.ndarray
        Predictive x-loadings.
    orthogonal_loadings: np.ndarray
        Orthogonal x-loadings.
    orthogonal_scores: np.ndarray
        Orthogonal x-scores.
    pls_: PLS
        The wrapped PLS model
    """

    def __init__(self, n_components=1, pls_components=2, *,
                 scale=True,
                 center=True,
                 flip=False,
                 max_iter=500,
                 tol=1e-06,
                 copy=True,
                 dtype=np.float64,
                 algorithm="OPLS",
                 deflation_mode=None):

        super().__init__(
            n_components=n_components,
            scale=scale,
            center=center,
            deflation_mode=deflation_mode,
            algorithm=algorithm,
            max_iter=max_iter,
            tol=tol,
            copy=copy,
            dtype=dtype,
            flip=flip
        )
        self.pls_components = pls_components

    def __str__(self):
        return f"{type(self).__name__}(n_components={self.n_components}, pls_components={self.pls_components})"

    def __repr__(self):
        return f"{type(self).__name__}(n_components={self.n_components}, pls_components={self.pls_components}, algorithm={repr(self.algorithm)}, deflation_mode={repr(self.deflation_mode)})"

    def fit(self, x: np.ndarray, y: np.ndarray):
        """
        Fit OPLS and PLS models.
        Parameters
        ----------
        x: np.ndarray
            Variable matrix with shape (n_samples, n_features)
        y: np.ndarray
            Dependent matrix with shape (n_samples, n_targets) or (n_samples,). For now only n_targets==1 is tested.

        Returns
        -------
        Fitted OPLS_PLS object (reference to self)

        Reference
        ---------
        [1] Trygg J, Wold S. Orthogonal projection on Latent Structure (OPLS).
            J Chemometrics. 2002, 16, 119-128.
        [2] Trygg J, Wold S. O2-PLS, a two-block (X-Y) latent variable
            regression (LVR) method with a integral OSC filter.
            J Chemometrics. 2003, 17, 53-64.
        [3] https://scikit-learn.org/stable/modules/cross_decomposition.html#cross-decomposition
        """

        n, xd = x.shape

        if (y.ndim == 1):
            yd = 1
        else:
            yd = y.shape[1]

        if (self.deflation_mode == "regression"):
            npc = min(n, xd)
        elif (self.deflation_mode == "canonical"):
            npc = min(n, xd, yd)
        else:
            raise ValueError(
                f"deflation_mode=\"{self.deflation_mode}\" is not supported. Supported values are \"canonical\" and \"regression\"")

        if (self.n_components+self.pls_components > npc):
            raise ValueError(
                f"Number of components is too large for X=shape{x.shape}, Y=shape{y.shape} and deflation_mode={self.deflation_mode}")

        super().fit(x, y)

        self._pls = PLS(
            n_components=self.pls_components,
            scale=self.scale,
            center=self.center,
            flip=self.flip,
            max_iter=self.max_iter,
            tol=self.tol,
            copy=self.copy,
            deflation_mode=self.deflation_mode,
        ).fit(*self.correct(x, y))
        self.pls_ = self._pls

        return self

    @overload
    def transform(self, X: np.ndarray, Y: np.ndarray, copy: bool = True) -> tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def transform(self, X: np.ndarray, Y: None = None, copy: bool = True) -> np.ndarray:
        ...

    def transform(self, X: np.ndarray, Y: Optional[np.ndarray] = None, copy: bool = True) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """Remove the orthogonal components and apply the dimension reduction.

        Parameters
        ----------
        X : np.ndarray
            shape(n_samples, n_features) coordinates to transform.
        Y : np.ndarray, default=None
            shape(n_samples, n_targets) targets to transform (optional)
        copy : bool, default=True
            Whether to copy `X` and `Y`, or perform in-place normalization.

        Returns
        -------
        x_scores : np.ndarray
            shape(n, pls_comp) x-scores.
        y_scores : np.ndarray
            shape(n, pls_comp) y-scores, only returned if Y is not None.
        """
        if (Y is None):
            x_filt = super().correct(X, copy=copy)
            return self._pls.transform(x_filt, copy=copy)
        x_filt, y_filt = super().correct(X, Y, copy=copy)
        return self._pls.transform(x_filt, y_filt, copy=copy)

    @overload
    def inverse_transform(self, X: np.ndarray, Y: None = None) -> np.ndarray:
        ...

    @overload
    def inverse_transform(self, X: np.ndarray, Y: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        ...

    def inverse_transform(self, X: np.ndarray, Y: Optional[np.ndarray] = None):
        """calculate inverse of the dimension reduction.

        Parameters
        ----------
        X : np.ndarray
            shape(n_samples, n_comp) scores to inverse transform.
        Y : np.ndarray, default=None
            shape(n_samples, n_comp) scores to inverse transform (optional)

        Returns
        -------
        x_scores : np.ndarray
            shape(n_samples, n_features) estimate of X-coordinates.
        y_scores : np.ndarray
            shape(n_samples, n_targets) estimate of y-targets, only returned if Y is not None.
        """
        return self._pls.inverse_transform(X, Y)

    @overload
    def transform_ortho(self, X: np.ndarray, Y: np.ndarray, *, copy: bool = True) -> tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def transform_ortho(self, X: np.ndarray, Y: None = None, *, copy: bool = True) -> np.ndarray:
        ...

    def transform_ortho(self, X: np.ndarray, Y: Optional[np.ndarray] = None, *,
                        copy: bool = True) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """Apply the dimension reduction to get the orthogonal components.

        Parameters
        ----------
        X : np.ndarray
            shape(n_samples, n_features) coordinates to transform.
        Y : np.ndarray, default=None
            shape(n_samples, n_targets) targets to transform (optional)
        copy : bool, default=True
            Whether to copy `X` and `Y`, or perform in-place normalization.

        Returns
        -------
        x_scores : np.ndarray
            shape(n_samples, n_comp) orthogonal x-scores.
        y_scores : np.ndarray
            shape(n_samples, n_comp) orthogonal y-scores OR unchanged Y if algorithm=="OPLS"
            only returned if Y is not None.
        """
        return super().transform(X, Y, copy=copy)

    @overload
    def inverse_transform_ortho(self, X: np.ndarray, Y: None = None) -> np.ndarray:
        ...

    @overload
    def inverse_transform_ortho(self, X: np.ndarray, Y: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        ...

    def inverse_transform_ortho(self, X: np.ndarray, Y: Optional[np.ndarray] = None) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """calculate inverse of the dimension reduction.

        Parameters
        ----------
        X : np.ndarray
            shape(n_samples, n_comp) scores to inverse transform.
        Y : np.ndarray, default=None
            shape(n_samples, n_comp) scores to inverse transform (optional)

        Returns
        -------
        x_scores : np.ndarray
            shape(n_samples, n_features) estimate of X-coordinates.
        y_scores : np.ndarray
            shape(n_samples, n_targets) estimate of y-targets OR unchanged Y if algorithm=="OPLS"
            only returned if Y is not None.
        """
        return super().inverse_transform(X, Y)

    def predict(self, X: np.ndarray, *, ndim: Optional[int] = None, copy: bool = True) -> np.ndarray:
        X_new = self.correct(X, copy=copy)
        return self.pls_.predict(X_new, ndim=ndim, copy=copy)

    def inverse_predict(self, Y: np.ndarray, *, ndim: Optional[int] = None, copy: bool = True) -> np.ndarray:
        return self.pls_.inverse_predict(Y, ndim=ndim, copy=copy)

    def score(self, X: np.ndarray, y: np.ndarray, sample_weight: Optional[np.ndarray] = None, *,
              ndim: Optional[int] = None,
              copy: bool = True) -> float:
        return self.pls_.score(*self.correct(X, y, copy=copy), sample_weight=sample_weight, ndim=ndim, copy=copy)



import numpy as np
from numpy import linalg as la

from typing import Optional


def center_scale_data(X: np.ndarray, Y: np.ndarray, center=True, scale=True) -> tuple[
        np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Centering and scaling of 2d arrays X and Y, along the first dimension.
    The input array will not be copied and the returned references should match the input

    Parameters
    ----------
    X,Y: np.ndarray
        2d arrays of shape(n,d), shape(m,f).
    center: bool, default=True
        Whether to do the mean centering. If False also scaling is not done.
    scale: bool, default=True
        Whether to scale the data to unit variance. Effectively divides the data with the
        respective (corrected) sample standard deviations.

    Returns
    -------
    X: np.ndarray
        The corrected X of shape (n,d). Should be a reference to the input X.
    Y: np.ndarray
        The corrected Y of shape (n,f). Should be a reference to the input Y.
    x_mean: np.ndarray
        shape(d) array of the means of X. If center=False, it is all zeroes.
    y_mean: np.ndarray
        shape(f) array of the means of Y. If center=False, it is all zeroes.
    x_std: np.ndarray
        shape(d) array of the standard deviations of X. Any zero values are replaced by 1.
        if center=False or scale=False it is all ones.
    y_std: np.ndarray
        shape(f) array of the standard deviations of Y. Any zero values are replaced by 1.
        if center=False or scale=False it is all ones.
    """
    if (center):
        x_mean = X.mean(axis=0)
        y_mean = Y.mean(axis=0)
        X -= x_mean
        Y -= y_mean
    else:
        x_mean = np.zeros(X.shape[1])
        y_mean = np.zeros(Y.shape[1])

    if (scale and center):
        x_std = X.std(axis=0, ddof=1)
        y_std = Y.std(axis=0, ddof=1)
        x_std = np.where(x_std != 0, x_std, 1.0)
        y_std = np.where(y_std != 0, y_std, 1.0)
        X /= x_std
        Y /= y_std
    else:
        x_std = np.ones(X.shape[1])
        y_std = np.ones(Y.shape[1])

    return X, Y, x_mean, y_mean, x_std, y_std


def nipals(x: np.ndarray, y: np.ndarray,
           tol: float = 1e-10,
           max_iter: int = 10000) -> tuple[
        np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Non-linear Iterative Partial Least Squares

    Parameters
    ----------
    x: np.ndarray
        Variable matrix with size n by d, where n number of samples,
        p number of variables.
    y: np.ndarray
        Dependent variable with size n by t.
    tol: float
        Tolerance for the convergence.
    max_iter: int
        Maximal number of iterations.

    Returns
    -------
    w: np.ndarray
        X-weights with size d by 1.
    c: np.ndarray
        Y-weight with size t by 1
    t: np.ndarray
        X-scores with size n by 1
    u: np.ndarray
        Y-scores with size n by 1.

    References
    ----------
    [1] Wold S, et al. PLS-regression: a basic tool of chemometrics.
        Chemometr Intell Lab Sys 2001, 58, 109–130.
    [2] Bylesjo M, et al. Model Based Preprocessing and Background
        Elimination: OSC, OPLS, and O2PLS. in Comprehensive Chemometrics.
    """
    u = y[:, 0]
    u = u[:, np.newaxis]
    i = 0
    d: np.floating | float = tol * 10
    while d > tol and i <= max_iter:
        w = (x.T @ u) / (u.T @ u)
        w /= la.norm(w)
        t = x @ w
        c = y.T @ t / (t.T @ t)
        u_new = y @ c / (c.T @ c)
        d = la.norm(u_new - u) / la.norm(u_new)
        u = u_new
        i += 1

    return w, c, t, u


def r2_score(y_true: np.ndarray, y_hat: np.ndarray, sample_weights: Optional[np.ndarray] = None) -> float:
    if (y_true.ndim == 1):
        y_true = y_true[:, None]
    if (y_hat.ndim == 1):
        y_hat = y_hat[:, None]
    if sample_weights is None:
        weights: float | np.ndarray = 1.0
    else:
        weights = sample_weights[:, None]

    numerator = (weights*(y_true - y_hat) ** 2).sum(axis=0, dtype=np.float64)
    y_mean = np.average(y_true, axis=0,
                        weights=sample_weights)
    denominator = (
        weights * (y_true - y_mean) ** 2
    ).sum(axis=0, dtype=np.float64)
    # Both zero: perfect prediction of zero variance
    both_zero = denominator == 0 * numerator == 0
    denominator[both_zero] = 1.0
    numerator[both_zero] = 1.0
    return np.mean(1-numerator/denominator)


def flip_scores_by_absolute_value(u: np.ndarray, v: np.ndarray):
    """Flips u and v such that the biggest absolute value in u will have a positive sign.

    Flipping makes the PLS model the same (up to machine precision) as with sklearn.

    The flipping is done in place.
    """
    biggest_abs_val_idx = np.argmax(np.abs(u))
    sign = np.sign(u[biggest_abs_val_idx])
    u *= sign
    v *= sign



#!/usr/bin/env python3
from typing import Literal, Optional, overload
import numpy as np
from numpy import linalg as la
from scipy.linalg import pinv

from .base_PLS import _PLS
from .utils import center_scale_data, nipals, flip_scores_by_absolute_value


class OPLS(
    _PLS
):
    """
    Orthogonal Projection on Latent Structure (O-PLS).

    Parameters
    ----------
    n_components : int, default=1
        Number of components to fit.
    scale : bool, default=True
        Whether to scale X and Y to unit variance
    center : bool, default=True
        Whether to centre X and Y to zero mean
    flip : bool, default=False
        Whether to flip the singular vectors for compatibility with different solvers.
        With flip=True and deflation_mode="regression" the first PLS component will be
        exactly (up to machine precision) the same as with sklearn.cross_decomposition.PLSRegression.
        Does not affect the results in any other meaningful way.
    max_iter : int, default=500
        Maximum number of NIPALS iterations.
    tol : float, default=1e-06
        Tolerance for stopping NIPALS iterations.
    copy : bool, default=True
        Whether to make copies of X and Y. False does not guarantee everything is in place, but
        True does guarantee copying.
    dtype : numpy.dtype, default=float64
        numpy dtype to cast input to. If None calculations will be done with whatever dtype the input arrays have.
    algorithm : str, default="OPLS"
        The algorithm to use. Acceptable values are "OPLS" and "O2PLS".
        NOTE: "O2PLS" is not yet well tested, and OPLS is only tested with univariate y.
    deflation_mode : str, default=None
        Whether to calculate y deflation with x_scores ("regression") or y_scores ("canonical").
        The latter only is reliable with n_components<=Y.shape[1].
        With OPLS algorithm this only changes the y-loadings, y-rotations and the final regressor, so
        it does not affect the corrected coordinates.
        With O2PLS algorithm "canonical" should be used.
        If None, "regression" is used for OPLS and "canonical" with O2PLS.

    Attributes
    ----------
    predictive_scores: np.ndarray
        Predictive x-scores.
    predictive_loadings: np.ndarray
        Predictive x-loadings.
    orthogonal_loadings: np.ndarray
        Orthogonal x-loadings.
    orthogonal_scores: np.ndarray
        Orthogonal x-scores.

    """

    def __init__(
        self, n_components=1, *,
        scale=True,
        center=True,
        flip=False,
        max_iter=500,
        tol=1e-06,
        copy=True,
        dtype=np.float64,
        algorithm="OPLS",
        deflation_mode=None,
    ):
        if (deflation_mode is None):
            if (algorithm == "OPLS"):
                deflation_mode = "regression"
            elif (algorithm == "O2PLS"):
                deflation_mode = "canonical"

        super().__init__(
            n_components=n_components,
            scale=scale,
            center=center,
            flip=flip,
            max_iter=max_iter,
            tol=tol,
            copy=copy,
            dtype=dtype,
            deflation_mode=deflation_mode
        )
        self.algorithm = algorithm

    def __str__(self):
        return f"{type(self).__name__}(n_components={self.n_components})"

    def __repr__(self):
        return f"{type(self).__name__}(n_components={self.n_components}, algorithm={repr(self.algorithm)}, deflation_mode={repr(self.deflation_mode)})"

    def fit(self, x: np.ndarray, y: np.ndarray):
        """
        Fit OPLS model.

        Parameters
        ----------
        x: np.ndarray
            Variable matrix with shape (n_samples, n_features)
        y: np.ndarray
            Dependent matrix with shape (n_samples, n_targets) or (n_samples,). For now only n_targets==1 is tested.

        Returns
        -------
        Fitted OPLS object (reference to self)

        Reference
        ---------
        [1] Trygg J, Wold S. Orthogonal projection on Latent Structure (OPLS).
            J Chemometrics. 2002, 16, 119-128.
        [2] Trygg J, Wold S. O2-PLS, a two-block (X-Y) latent variable
            regression (LVR) method with a integral OSC filter.
            J Chemometrics. 2003, 17, 53-64.
        [3] https://scikit-learn.org/stable/modules/cross_decomposition.html#cross-decomposition
        """
        X, Y = self.validate_input(x, y)
        if (Y.ndim == 1):
            Y = Y[:, np.newaxis]

        n, xd = X.shape
        _, yd = Y.shape

        if (self.deflation_mode not in ("canonical", "regression")):
            raise ValueError(
                f"deflation_mode=\"{self.deflation_mode}\" is not supported. Supported values are \"canonical\" and \"regression\"")

        if (self.algorithm == "OPLS"):
            npc = min(n, xd)
        elif (self.algorithm == "O2PLS"):
            npc = min(n, xd, yd)
        else:
            raise ValueError(
                f"algorithm=\"{self.algorithm}\" is not supported. Supported values are \"OPLS\" and \"O2PLS\"")

        if self.n_components > npc:
            raise ValueError(
                f"Number of components is too large for X=shape{X.shape}, Y=shape{Y.shape} and deflation_mode={self.deflation_mode}")

        n_comp = self.n_components

        if (self.n_components <= 0):
            raise ValueError(
                f"n_components should be positive nonzero integer, is {self.n_components}")

        X, Y, self._x_mean, self._y_mean, self._x_std, self._y_std = center_scale_data(
            X, Y, scale=self.scale, center=self.center)
        self.intercept_ = self._y_mean

        #  Variable              | name       |    variable in sklearn user guide
        W = np.empty((xd, n_comp))  # X-weights  |     U
        C = np.empty((yd, n_comp))  # Y-weights  |     V
        T = np.empty((n, n_comp))   # X-scores   |     Xi
        U = np.empty((n, n_comp))   # Y-scores   |     Omega
        P = np.empty((xd, n_comp))  # X-loadings |     Gamma
        Q = np.empty((yd, n_comp))  # Y-loadings |     Delta

        # Orthogonal variables
        Wortho = np.empty((xd, n_comp))  # X-weights
        Tortho = np.empty((n, n_comp))   # X-scores
        Portho = np.empty((xd, n_comp))  # X-loadings
        if (self.algorithm == "O2PLS"):
            Cortho = np.empty((yd, n_comp))  # Y-weights
            Uortho = np.empty((n, n_comp))   # Y-scores
            Qortho = np.empty((yd, n_comp))  # Y-loadings

        # Y_eps = np.finfo(Y.dtype).eps

        for k in range(n_comp):
            # Replace columns that are all close to zero with zeros
            # Y_mask = np.all(np.abs(Y) < 10 * Y_eps, axis=0)
            # Y[:, Y_mask] = 0.0

            # Run nipals to get first singular vectors
            w, c, t, u = nipals(X, Y, tol=self.tol, max_iter=self.max_iter)

            if (self.flip):
                # Flip for consistency across solvers
                flip_scores_by_absolute_value(w, c)
                # recalculate scores after flip
                t = X @ w
                u = Y @ c / (c.T @ c)

            # Regress p to minimize error in Xhat = t p^T
            p = (X.T @ t) / (t.T @ t)

            # find orthogonal weights
            w_ortho = p-((w.T @ p) / (w.T @ w))*w
            w_ortho /= la.norm(w_ortho)

            # with the orthogonal weights, calculate orthogonal scores and loadings
            t_ortho = (X @ w_ortho) / (w_ortho.T @ w_ortho)
            p_ortho = (X.T @ t_ortho) / (t_ortho.T @ t_ortho)

            # orthogonal deflation of X
            X -= t_ortho @ p_ortho.T

            if (self.deflation_mode == "canonical"):
                # Regress q to minimize error in Yhat = u q^T
                q = (Y.T @ u) / (u.T @ u)
            elif (self.deflation_mode == "regression"):
                # In regression mode only x score (t) is used
                # Regress q to minimize error in Yhat = t q^T
                q = (Y.T @ t) / (t.T @ t)

            # In O2PLS we also calculate y-loadings and deflate y
            if (self.algorithm == "O2PLS"):
                # find orthogonal y-weights
                c_ortho = q-((c.T @ q) / (c.T @ c))*c
                c_ortho /= la.norm(c_ortho)

                # with the orthogonal weights, calculate orthogonal scores and loadings
                u_ortho = (Y @ c_ortho) / (c_ortho.T @ c_ortho)
                if (self.deflation_mode == "canonical"):
                    q_ortho = (Y.T @ u_ortho) / (u_ortho.T @ u_ortho)
                    # deflate y
                    Y -= u_ortho @ q_ortho.T
                elif (self.deflation_mode == "regression"):
                    # In regression mode only x score (t_ortho) is used
                    q_ortho = (Y.T @ t_ortho) / (t_ortho.T @ t_ortho)
                    # deflate y
                    Y -= t_ortho @ q_ortho.T

            W[:, k] = w.squeeze(axis=1)
            U[:, k] = u.squeeze(axis=1)
            C[:, k] = c.squeeze(axis=1)
            T[:, k] = t.squeeze(axis=1)
            P[:, k] = p.squeeze(axis=1)
            Q[:, k] = q.squeeze(axis=1)

            Wortho[:, k] = w_ortho.squeeze(axis=1)
            Tortho[:, k] = t_ortho.squeeze(axis=1)
            Portho[:, k] = p_ortho.squeeze(axis=1)
            if (self.algorithm == "O2PLS"):
                Uortho[:, k] = u_ortho.squeeze(axis=1)
                Cortho[:, k] = c_ortho.squeeze(axis=1)
                Qortho[:, k] = q_ortho.squeeze(axis=1)

        self._x_weights = W
        self._y_weights = C
        self._x_scores = T
        self._y_scores = U
        self._x_loadings = P
        self._y_loadings = Q

        self._Wortho = Wortho
        self._Tortho = Tortho
        self._Portho = Portho

        self.x_rotations_ = W @ pinv(P.T @ W, check_finite=False)
        self.y_rotations_ = C @ pinv(Q.T @ C, check_finite=False)

        # orthogonal x rotations
        self._Rortho = Wortho @ pinv(Portho.T @ Wortho, check_finite=False)

        if (self.algorithm == "O2PLS"):
            self._Uortho = Uortho
            self._Cortho = Cortho
            self._Qortho = Qortho
            # orthogonal y rotations
            self._Sortho = Cortho @ pinv(Qortho.T @ Cortho, check_finite=False)

        # "expose" all the weights, scores and loadings
        self.x_weights_ = self._x_weights
        self.y_weights_ = self._y_weights
        self.x_scores_ = self._x_scores
        self.y_scores_ = self._y_scores
        self.x_loadings_ = self._x_loadings
        self.y_loadings_ = self._y_loadings

        self.fitted = True
        return self

    def predict(self, X: np.ndarray, *, ndim: Optional[int] = None, copy: bool = True) -> np.ndarray:
        raise NotImplementedError("predict not available with OPLS")

    def score(self, X: np.ndarray, y: np.ndarray, sample_weight: Optional[np.ndarray] = None, *,
              ndim: Optional[int] = None,
              copy: bool = True) -> float:
        raise NotImplementedError("score not available with OPLS")

    def inverse_predict(self, Y, *, ndim: Optional[int] = None, copy: bool = True):
        raise NotImplementedError("inverse_predict not available with OPLS")

    @overload
    def transform(self, X: np.ndarray, Y: np.ndarray, *, copy: bool = True) -> tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def transform(self, X: np.ndarray, Y: None = None, *, copy: bool = True) -> np.ndarray:
        ...

    def transform(self, X: np.ndarray, Y: Optional[np.ndarray] = None, *, copy: bool = True) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """Apply the dimension reduction to get the orthogonal scores.

        Parameters
        ----------
        X : np.ndarray
            shape(n_samples, n_features) coordinates to transform.
        Y : np.ndarray, default=None
            shape(n_samples, n_targets) targets to transform (optional)
        copy : bool, default=True
            Whether to copy `X` and `Y`, or perform in-place normalization.

        Returns
        -------
        x_scores : np.ndarray
            shape(n, n_comp) orthogonal x-scores.
        y_scores : np.ndarray
            shape(n, n_comp) orthogonal y-scores OR unchanged Y if algorithm=="OPLS"
            only returned if Y is not None.
        """
        self.check_is_fitted()
        X = self._validate_array(X, variable_name="Y")
        # Normalize
        X -= self._x_mean
        X /= self._x_std
        # calculate scores
        x_scores = X @ self._Rortho
        if Y is not None:
            if (self.algorithm == "OPLS"):
                return x_scores, Y
            Y = self._validate_array(
                Y, variable_name="Y", ensure_2d=False, copy=copy
            )
            if Y.ndim == 1:
                Y = Y[:, np.newaxis]
            Y -= self._y_mean
            Y /= self._y_std
            y_scores = Y @ self._Sortho
            return x_scores, y_scores

        return x_scores

    @overload
    def inverse_transform(self, X: np.ndarray, Y: None = None) -> np.ndarray:
        ...

    @overload
    def inverse_transform(self, X: np.ndarray, Y: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        ...

    def inverse_transform(self, X: np.ndarray, Y: Optional[np.ndarray] = None) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """calculate inverse of the dimension reduction.

        Parameters
        ----------
        X : np.ndarray
            shape(n_samples, n_comp) scores to inverse transform.
        Y : np.ndarray, default=None
            shape(n_samples, n_comp) scores to inverse transform (optional)

        Returns
        -------
        x_scores : np.ndarray
            shape(n_samples, n_features) estimate of X-coordinates.
        y_scores : np.ndarray
            shape(n_samples, n_targets) estimate of y-targets OR unchanged Y if algorithm=="OPLS"
            only returned if Y is not None.
        """
        self.check_is_fitted()
        X = self._validate_array(X, variable_name="X")
        # calculate scores
        x_new = X @ self._Portho.T
        x_new *= self._x_std
        x_new += self._x_mean

        if Y is not None:
            if (self.algorithm == "OPLS"):
                return x_new, Y
            Y = self._validate_array(Y, variable_name="Y")
            y_new = Y @ self._Qortho.T
            y_new *= self._y_std
            y_new += self._y_mean
            return x_new, y_new

        return x_new

    @overload
    def correct(self, X: np.ndarray, y: None = None, *,
                copy: bool = True,
                return_ortho: Literal[False] = False) -> np.ndarray:
        ...

    @overload
    def correct(self, X: np.ndarray, y: None = None, *,
                copy: bool = True,
                return_ortho: Literal[True] = True) -> tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def correct(self, X: np.ndarray, y: np.ndarray, *,
                copy: bool = True,
                return_ortho: Literal[False] = False) -> tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def correct(self, X: np.ndarray, y: np.ndarray, *,
                copy: bool = True,
                return_ortho: Literal[True] = True) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        ...

    @overload
    def correct(self, X: np.ndarray, y: Optional[np.ndarray] = None, *,
                copy: bool = True,
                return_ortho: bool = False) -> np.ndarray | tuple[np.ndarray, np.ndarray] |\
            tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        ...

    def correct(self, X: np.ndarray, y: Optional[np.ndarray] = None, *,
                copy: bool = True,
                return_ortho: bool = False) -> np.ndarray | tuple[np.ndarray, ...]:
        """
        Remove orthogonal components from X (and possibly y)

        Parameters
        ----------
        X: np.ndarray
            Data matrix with shape(n_samples, n_features)
        y: np.ndarray | None
            Data matrix shape(n_samples) or shape(n_samples, n_targets).
            If None, only X is corrected and returned.
            If the algorithm was OPLS, the corrected y is simply a copy of y.
        copy: bool
            Wether to work on and return copies of data. Default is True.
        return_ortho: bool
            Return orthogonal components of X (and possibly y). Default is False.

        Returns
        -------
        x_new: np.ndarray
            Corrected data, shape(n_samples, n_features).
        x_ortho: np.ndarray
            Orthogonal score, shape(n_samples, n_features). Returned if return _ortho=True.
        y_new: np.ndarray
            Corrected data, shape(n_samples, n_targets). Returned if y is not None.
        y_ortho: np.ndarray
            Orthogonal score, shape(n_samples, n_targets). Returned if y is not None and return _ortho=True.
        """
        self.check_is_fitted()
        if (not y is None):
            x_new, y_new = self.validate_input(X, y, copy=copy)
        else:
            x_new = self._validate_array(X, variable_name="X", copy=copy)
        x_new -= self._x_mean
        x_new /= self._x_std
        n, xd = x_new.shape

        if (xd != self._x_loadings.shape[0]):
            raise ValueError("Dimension mismatch in X, "
                             f"the model has been trained with nd={self._x_loadings.shape[0]},"
                             f" but X is shape{X.shape}")

        if (not y is None):
            if (len(y_new.shape) == 1):
                y_new = y_new[:, np.newaxis]
            if (y_new.shape[1] != self._y_loadings.shape[0]):
                raise ValueError("Dimension mismatch in y, "
                                 f"the model has been trained with nf={self._y_loadings.shape[0]},"
                                 f" but y is shape{y.shape}")

        correct_y = (not y is None) and self.algorithm == "O2PLS"

        if (correct_y):
            y_new -= self._y_mean
            y_new /= self._y_std

        # The division is not needed as Wortho should be orthonormal, so Wortho.T @ Wortho = I
        # If the normalisation is removed, the division will be needed
        T_ortho_new = (x_new @ self._Rortho)
        # / np.diag(self._Wortho.T @ self._Wortho)

        x_ortho = T_ortho_new @ self._Portho.T

        x_new -= x_ortho
        x_new *= self._x_std
        x_new += self._x_mean

        if correct_y:
            # The division is not needed as Cortho should be orthonormal, so Cortho.T @ Cortho = I
            U_ortho_new = (y_new @ self._Sortho)
            # / np.diag(self._Cortho.T @ self._Cortho)

            y_ortho = U_ortho_new @ self._Qortho.T

            y_new -= y_ortho

            y_new *= self._y_std
            y_new += self._y_mean

        if not return_ortho:
            if (not y is None):
                if (len(y.shape) == 1):
                    return x_new, y_new.squeeze(axis=1)
                return x_new, y_new
            return x_new

        x_ortho *= self._x_std
        x_ortho += self._x_mean
        if (not y is None):
            if (correct_y):
                y_ortho *= self._y_std
                y_ortho += self._y_mean
            else:
                y_ortho = np.zeros_like(y_new)
            if (len(y.shape) == 1):
                return x_new, x_ortho, y_new.squeeze(axis=1), y_ortho.squeeze(axis=1)
            return x_new, x_ortho, y_new, y_ortho
        return x_new, x_ortho

    @property
    def predictive_scores(self) -> np.ndarray:
        """ Orthogonal loadings. """
        return self._x_scores

    @property
    def predictive_loadings(self) -> np.ndarray:
        """ Predictive loadings. """
        return self._x_loadings

    @property
    def weights_y(self) -> np.ndarray:
        """ y scores. """
        return self._y_weights

    @property
    def orthogonal_loadings(self) -> np.ndarray:
        """ Orthogonal loadings. """
        return self._Portho

    @property
    def orthogonal_scores(self) -> np.ndarray:
        """ Orthogonal scores. """
        return self._Tortho



from typing import Optional
import numpy as np


class _PLS:
    def __init__(self, n_components=2, *,
                 scale=True,
                 center=True,
                 flip=False,
                 max_iter=500,
                 tol=1e-06,
                 copy=True,
                 dtype=np.float64,
                 deflation_mode="regression"):
        self.n_components = n_components
        self.scale = scale
        self.center = center
        self.flip = flip
        self.max_iter = max_iter
        self.tol = tol
        self.copy = copy
        self.dtype = dtype
        self.deflation_mode = deflation_mode
        self.fitted = False

    def _validate_array(self, x: np.ndarray, *,
                        variable_name="x",
                        copy: Optional[bool] = None,
                        ensure_min_samples: int = 2,
                        ensure_2d: bool = True):
        if not isinstance(x, np.ndarray):
            raise ValueError(
                f"{variable_name} should be numpy.ndarray, is {type(x)}"
            )
        if (ensure_2d):
            if x.ndim != 2:
                raise ValueError(
                    f"{variable_name}.shape={x.shape}, should be 2d when ensure_2d=True"
                )
        else:
            if x.ndim not in (1, 2):
                raise ValueError(
                    f"{variable_name}.shape={x.shape}, should be 1d or 2d when ensure_2d=False"
                )

        if (x.shape[0] < ensure_min_samples):
            raise ValueError(
                f"{variable_name}.shape[0]={x.shape[0]}, should be at least {ensure_min_samples}"
            )

        if (copy is None):
            copy = self.copy

        dtype = self.dtype
        if (dtype is None):
            dtype = x.dtype

        return np.array(x, dtype=dtype, copy=copy)

    def validate_input(self, x: np.ndarray, y: np.ndarray, *,
                       copy: Optional[bool] = None,
                       ensure_min_samples: int = 2,
                       allow_univariate_y=True):
        x = self._validate_array(x, variable_name="x",
                                 copy=copy,
                                 ensure_min_samples=ensure_min_samples,
                                 ensure_2d=True)
        y = self._validate_array(y, variable_name="y",
                                 copy=copy,
                                 ensure_min_samples=ensure_min_samples,
                                 ensure_2d=not allow_univariate_y)
        if (x.shape[0] != y.shape[0]):
            raise ValueError(
                f"First dimension of x and y do not match, {x.shape[0]}!={y.shape[0]}"
            )

        return x, y

    def check_is_fitted(self):
        if not self.fitted:
            raise ValueError(
                f"You should run {type(self)}.fit before any prediction functions."
            )



from typing import Optional, overload, Literal
import numpy as np
from scipy.linalg import pinv

from .base_PLS import _PLS
from .utils import center_scale_data, nipals, r2_score, flip_scores_by_absolute_value


class PLS(
    _PLS
):
    """
    Parameters
    ----------
    n_components : int, default=2
        Number of components to fit.
    scale : bool, default=True
        Whether to scale X and Y to unit variance
    center : bool, default=True
        Whether to centre X and Y to zero mean
    flip : bool, default=False
        Whether to flip the singular vectors for compatibility with different solvers.
        With flip=True and deflation_mode="regression" the PLS model will be the same
        (up to machine precision) as with sklearn.cross_decomposition.PLSRegression.
        With flip=False and univariate y, the x_scores_ are guaranteed to be positively
        correlated to y.
    max_iter : int, default=500
        Maximum number of NIPALS iterations.
    tol : float, default=1e-06
        Tolerance for stopping NIPALS iterations.
    copy : bool, default=True
        Whether to make copies of X and Y. False does not guarantee calculations are in place, but
        True does guarantee copying.
    dtype : numpy.dtype, default=float64
        numpy dtype to cast input to. If None calculations will be done with whatever dtype the input arrays have.
    deflation_mode : str, default=None
        Whether to calculate y deflation with x_scores ("regression") or y_scores ("canonical").
        The latter only is reliable with n_components<=Y.shape[1]. The first will make this the same as
        sklearn.cross_decomposition.PLSRegression and the latter the same as
        sklearn.cross_decomposition.PLSCanonical.
        In either case the first PLS component is the same.
    """

    def __str__(self):
        return f"{type(self).__name__}(n_components={self.n_components})"

    def __repr__(self):
        return f"{type(self).__name__}(n_components={self.n_components}, deflation_mode={repr(self.deflation_mode)})"

    def fit(self, x: np.ndarray, y: np.ndarray):
        """
        Fit PLS model.
        Parameters
        ----------
        x: np.ndarray
            Variable matrix with size n samples by n_features variables.
        y: np.ndarray
            Dependent matrix with size n samples by n_targets, or a vector. For now only t==1 is implemented.
        n_comp: int
            Number of components, default is None, which indicates that
            largest dimension which is smaller value between n and p
            will be used.
        Returns
        -------
        PLS object
        Reference
        ---------
        [1] Trygg J, Wold S. Orthogonal projection on Latent Structure (OPLS).
            J Chemometrics. 2002, 16, 119-128.
        [2] Trygg J, Wold S. O2-PLS, a two-block (X-Y) latent variable
            regression (LVR) method with a integral OSC filter.
            J Chemometrics. 2003, 17, 53-64.
        [3] https://scikit-learn.org/stable/modules/cross_decomposition.html#cross-decomposition
        """
        X, Y = self.validate_input(x, y)
        if (Y.ndim == 1):
            Y = Y[:, np.newaxis]

        n, xd = X.shape
        _, yd = Y.shape

        if (self.deflation_mode == "regression"):
            npc = min(n, xd)
        elif (self.deflation_mode == "canonical"):
            npc = min(n, xd, yd)
        else:
            raise ValueError(
                f"deflation_mode=\"{self.deflation_mode}\" is not supported. Supported values are \"canonical\" and \"regression\"")

        if self.n_components > npc:
            raise ValueError(
                f"Number of components is too large for X=shape{X.shape}, Y=shape{Y.shape} and deflation_mode={self.deflation_mode}. Maximum value is {npc}.")

        if (self.n_components <= 0):
            raise ValueError(
                f"n_components should be positive nonzero integer, is {self.n_components}")

        n_comp = self.n_components

        X, Y, self._x_mean, self._y_mean, self._x_std, self._y_std = center_scale_data(
            X, Y, center=self.center, scale=self.scale)
        self.intercept_ = self._y_mean

        #  Variable            | name       |    variable in sklearn user guide
        W = np.empty((xd, n_comp))  # X-weights  |     U
        C = np.empty((yd, n_comp))  # Y-weights  |     V
        T = np.empty((n,  n_comp))  # X-scores   |     Xi
        U = np.empty((n,  n_comp))  # Y-scores   |     Omega
        P = np.empty((xd, n_comp))  # X-loadings |     Gamma
        Q = np.empty((yd, n_comp))  # Y-loadings |     Delta

        all_coefs = np.empty((n_comp, yd, xd))

        # Y_eps = np.finfo(Y.dtype).eps

        for k in range(n_comp):
            # Replace columns that are all close to zero with zeros
            # Y_mask = np.all(np.abs(Y) < 10 * Y_eps, axis=0)
            # Y[:, Y_mask] = 0.0

            # Run nipals to get first singular vectors
            w, c, t, u = nipals(X, Y, tol=self.tol, max_iter=self.max_iter)

            if (self.flip):
                # Flip for consistency across solvers
                flip_scores_by_absolute_value(w, c)
                # recalculate scores after flip
                t = X @ w
                u = Y @ c / (c.T @ c)

            # Regress p to minimize error in Xhat = t p^T
            p = (X.T @ t) / (t.T @ t)
            # deflation of X
            X -= t @ p.T

            if (self.deflation_mode == "canonical"):
                # Regress q to minimize error in Yhat = u q^T
                q = (Y.T @ u) / (u.T @ u)
                # deflate y
                Y -= u @ q.T
            elif (self.deflation_mode == "regression"):
                # In regression mode only x score (t) is used
                # Regress q to minimize error in Yhat = t q^T
                q = (Y.T @ t) / (t.T @ t)
                # deflate y
                Y -= t @ q.T

            W[:, k] = w.squeeze(axis=1)
            U[:, k] = u.squeeze(axis=1)
            C[:, k] = c.squeeze(axis=1)
            T[:, k] = t.squeeze(axis=1)
            P[:, k] = p.squeeze(axis=1)
            Q[:, k] = q.squeeze(axis=1)

            _x_rot = W[:, :k+1] @ pinv(P[:, :k+1].T @
                                       W[:, :k+1], check_finite=False)
            _y_rot = C[:, :k+1] @ pinv(Q[:, :k+1].T @
                                       C[:, :k+1], check_finite=False)
            coef = _x_rot @ Q[:, :k+1].T
            coef *= self._y_std

            all_coefs[k] = coef.T

        self._x_weights = W
        self._y_weights = C
        self._x_scores = T
        self._y_scores = U
        self._x_loadings = P
        self._y_loadings = Q

        self.x_rotations_ = W @ pinv(P.T @ W, check_finite=False)
        self.y_rotations_ = C @ pinv(Q.T @ C, check_finite=False)

        self._coef_ = self.x_rotations_ @ Q.T
        self._coef_ *= self._y_std
        self._coef_ = self._coef_.T

        self._all_coefs = all_coefs

        self.coef_ = self._coef_

        # "expose" all the weights, scores and loadings
        self.x_weights_ = self._x_weights
        self.y_weights_ = self._y_weights
        self.x_scores_ = self._x_scores
        self.y_scores_ = self._y_scores
        self.x_loadings_ = self._x_loadings
        self.y_loadings_ = self._y_loadings

        self.fitted = True
        return self

    def score(self, X: np.ndarray, y: np.ndarray, sample_weight: Optional[np.ndarray] = None, *,
              ndim: Optional[int] = None,
              copy: bool = True) -> float:
        """Predict targets of given samples.
        Parameters
        ----------
        X : array-like of shape(n_samples, n_features)
            Samples.
        y : array-like of shape shape (n_samples) or (n_samples, n_targets)
            True y-values
        ndim : int|None, default None
            Number of PLS dimension to use for the prediction. None uses all.
        copy : bool, default=True
            Whether to copy `X` and `Y`, or perform in-place normalization.
            If `None`, uses value of self.copy.
        Returns
        -------
        r2_score : float
            Returns r2 score of predictions.
        """
        y = self._validate_array(y, variable_name="y", ensure_2d=False)
        if (y.shape[0] != X.shape[0]):
            raise ValueError(
                f"Inconsistent lengths between X and y ({X.shape=}, {y.shape=})"
            )
        yd = 1 if y.ndim == 1 else y.shape[1]
        if (yd != self.y_loadings_.shape[0]):
            raise ValueError(
                f"Wrong number of features in y ({y.shape=}, should be ({X.shape[0], self.y_loadings_.shape[0]}))"
            )
        if (sample_weight is not None and sample_weight.shape != (y.shape[0],)):
            sample_weight = self._validate_array(sample_weight,
                                                 variable_name="sample_weight",
                                                 ensure_2d=False)
            raise ValueError(
                f"Inconsistent lengths between sample_weight and y ({sample_weight.shape=}, {y.shape=})"
            )
        y_pred = self.predict(X, ndim=ndim, copy=copy)
        return r2_score(y, y_pred, sample_weights=sample_weight)

    @overload
    def transform(self, X: np.ndarray, y: np.ndarray, *, copy: bool = True) -> tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def transform(self, X: np.ndarray, y: None = None, *, copy: bool = True) -> np.ndarray:
        ...

    def transform(self, X: np.ndarray, y: Optional[np.ndarray] = None, *, copy=True) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """Predict latent space of given samples.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            X-samples.
        y : array-like of shape (n_samples) or (n_samples, n_targets), default=None
            Y-samples. Ignored if None.
        copy : bool, default=True
            Whether to copy `X` and `Y`, or perform in-place normalization.
            If `None`, uses value of self.copy.
        Returns
        -------
        x_scores : ndarray of shape (n_samples, n_features)
            Returns predicted latent space of X.
        y_scores : ndarray of shape (n_samples, n_targets), only if y is not None
            Returns predicted latent space of y.
        Notes
        -----
        This call requires the estimation of a matrix of shape
        `(n_features, n_targets)`, which may be an issue in high dimensional
        space.
        """
        self.check_is_fitted()
        if (y is not None):
            X, y = self.validate_input(X, y, copy=copy)
        else:
            X = self._validate_array(X, variable_name="X", copy=copy)
        X -= self._x_mean
        X /= self._x_std

        x_scores = X @ self.x_rotations_

        if (y is not None):
            y -= self._y_mean
            y /= self._y_std
            y_scores = y @ self.y_rotations_
            return x_scores, y_scores

        return x_scores

    @overload
    def inverse_transform(self, x_scores: np.ndarray, y_scores: None = None) -> np.ndarray:
        ...

    @overload
    def inverse_transform(self, x_scores: np.ndarray, y_scores: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        ...

    def inverse_transform(self, x_scores: np.ndarray, y_scores: Optional[np.ndarray] = None) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        """Predict latent space of given samples.
        Parameters
        ----------
        x_scores : array-like of shape (n_samples, n_components)
            X-scores (latent space).
        y_scores : array-like of shape(n_samples, n_components), default=None
            y-scores (latent space). Ignored if None.
        copy : bool, default=True
            Whether to copy `X` and `Y`, or perform in-place normalization.
            If `None`, uses value of self.copy.
        Returns
        -------
        X_hat : ndarray of shape (n_samples, n_features)
            Returns predicted real space X.
        y_hat : ndarray of shape (n_samples, n_targets), only if y_scores is not None
            Returns predicted real space y.
        Notes
        -----
        This call requires the estimation of a matrix of shape
        `(n_features, n_targets)`, which may be an issue in high dimensional
        space.
        """
        self.check_is_fitted()
        if (y_scores is not None):
            x_scores, y_scores = self.validate_input(x_scores,
                                                     y_scores,
                                                     copy=False)
        else:
            x_scores = self._validate_array(x_scores,
                                            variable_name="X",
                                            copy=False)

        X_hat = x_scores @ self.x_loadings_.T
        X_hat *= self._x_std
        X_hat += self._x_mean

        if (y_scores is not None):
            y_hat = y_scores @ self.y_loadings_.T
            y_hat *= self._y_std
            y_hat += self._y_mean
            return X_hat, y_hat

        return X_hat

    def predict(self, X: np.ndarray, *, ndim: Optional[int] = None, copy=True) -> np.ndarray:
        """Predict targets of given samples.
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Samples.
        ndim : int|None, default None
            Number of PLS dimension to use for the prediction. None uses all
        copy : bool, default=True
            Whether to copy `X` and `Y`, or perform in-place normalization.
            If `None`, uses value of self.copy.
        Returns
        -------
        y_pred : ndarray of shape (n_samples, n_targets)
            Returns predicted values.
        Notes
        -----
        This call requires the estimation of a matrix of shape
        `(n_features, n_targets)`, which may be an issue in high dimensional
        space.
        """
        if (ndim is None):
            ndim = self.n_components

        self.check_is_fitted()

        if (ndim > self.n_components):
            raise ValueError(
                f"ndim is bigger than the number of components this object was trained with")

        X = self._validate_array(X, variable_name="X", copy=copy)
        # Normalize
        X -= self._x_mean
        X /= self._x_std

        Ypred = X @ self._all_coefs[ndim-1].T
        return Ypred + self.intercept_

    def inverse_predict(self, Y: np.ndarray, *, ndim: Optional[int] = None, copy=True) -> np.ndarray:
        """Predict samples of given targets.
        With univariate y, this is a great way to visualize the final regression model, as this is just a linear interpolation of the coefficient vector
        along the given y-coordinates. For example:

        >>> x_interp = pls.inverse_predict(np.linspace(y.min(), y.max(), 101))

        With multivariate y this can still be used to generate interpolated structures, but it becomes a more complex combination of the coefficient vectors.
        In such a case it might be more meaningful to manually interpolate along each of the n_components coefficient vector individually. 

        Parameters
        ----------
        Y : np.ndarray
            Array of shape(n_samples) or shape(n_samples, n_targets) targets. In the first case this is done as
            linear interpolation along the coefficient vector. In the latter the pseudo inverse of the coeficient matrix is calculated.
            When n_targets=1 these two methods are equal (up to machine precision).

        copy : bool, default=True
            Whether to copy `X` and `Y`, or perform in-place normalization.

        Returns
        -------
        X_pred : np.ndarray
            shape (n_samples, n_features)
            Returns predicted values.

        """
        self.check_is_fitted()

        if (ndim is None):
            ndim = self.n_components

        if (ndim > self.n_components):
            raise ValueError(
                f"ndim is bigger than the number of components this object was trained with")

        Y = self._validate_array(
            Y, variable_name="Y", ensure_2d=False, copy=copy
        )
        # Center the Y values. _coef_ already has
        Y -= self.intercept_

        if Y.ndim == 1:
            # This is technically equal to the below with univariate y, but doesn't require the pseudo inversing
            scaledcoef = self._all_coefs[ndim-1] / \
                (self._all_coefs[ndim-1]**2).sum()
            X_pred = Y[:, np.newaxis] * scaledcoef
        else:
            invcoef = pinv(self._all_coefs[ndim-1]).T
            X_pred = Y @ invcoef

        X_pred *= self._x_std
        X_pred += self._x_mean
        return X_pred



"""
MD simulation wrappers for the PLS/OPLS objects in OPLS.py
"""
from typing import TYPE_CHECKING, Literal, Optional, overload
import numpy as np

try:
    from MDAnalysis import Universe, AtomGroup  # type: ignore
except ImportError as e:
    _reason = e
    if (not TYPE_CHECKING):
        Universe = None
        AtomGroup = None

from .OPLS import OPLS
from .PLS import PLS
from .OPLS_PLS import OPLS_PLS

# A dirty hack to make the super functions be loaded only when type checking
if TYPE_CHECKING:
    _Base_PLS = PLS
else:
    _Base_PLS = object


class _MD_PLS_WRAPPER(_Base_PLS):
    """
    A wrapper class to wrap PLS object to accept MDAnalysis universes and atomg groups,
    as well as unflattened MD trajectories. Should NOT be instanciated as such, but through the PLS_MD class.
    The final class needs to first inherit this class, and only then the class which is wrapped.
    """
    # This "Mixin" class is used instead of inheritance to be able to use he same wrappers for PLS and OPLS

    def _get_dims(self, crd: np.ndarray | Universe | AtomGroup):
        if (type(crd) == Universe or type(crd) == AtomGroup):
            self.natoms = crd.atoms.n_atoms
            self.ndim = 3
        else:
            if (crd.ndim == 2):
                self.ndim = 1
            elif (crd.ndim == 3):
                self.ndim = crd.shape[2]
            else:
                raise ValueError(
                    f"X is {crd.ndim} dimensional, should be 2 or 3")
            self.natoms = crd.shape[1]

    def _from_crd(self, crd: np.ndarray | Universe | AtomGroup) -> np.ndarray:
        if (type(crd) == Universe):
            u = crd
            sel = crd.atoms
        elif (type(crd) == AtomGroup):
            u = crd.universe
            sel = crd
        else:
            if (crd.ndim == 2):
                return crd
            if (crd.ndim != 3):
                raise ValueError(f"X is {crd.ndim} dimensional, "
                                 "should be 2 or 3")

            if (crd.shape[1:] != (self.natoms, self.ndim)):
                raise ValueError("Wrong shape in input, the model was trained "
                                 f"with shape{(-1, self.natoms, self.ndim)}, "
                                 f"input is {crd.shape}")

            return crd.reshape((crd.shape[0], self.natoms*self.ndim))

        if (sel.n_atoms != self.natoms):
            raise ValueError("Wrong number of atoms, the model was trained "
                             f"with {self.natoms}, input has {sel.n_atoms}")

        X = np.empty((len(u.trajectory), self.natoms*self.ndim))
        for i, ts in enumerate(u.trajectory):
            X[i] = sel.positions.ravel()
        return X

    def _to_crd(self, X: np.ndarray) -> np.ndarray:
        return X.reshape(X.shape[0], self.natoms, self.ndim)

    def fit(self, crd: np.ndarray | Universe | AtomGroup, y: np.ndarray):
        self._get_dims(crd)
        return super().fit(self._from_crd(crd), y)

    @overload
    def transform(self,
                  crd: np.ndarray | Universe | AtomGroup,
                  Y: np.ndarray,
                  *,
                  copy: bool = True) -> tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def transform(self,
                  crd: np.ndarray | Universe | AtomGroup,
                  Y: None = None,
                  *,
                  copy: bool = True) -> np.ndarray:
        ...

    def transform(self,
                  crd: np.ndarray | Universe | AtomGroup,
                  Y: Optional[np.ndarray] = None,
                  *,
                  copy: bool = True):
        X = self._from_crd(crd)
        return super().transform(X, Y, copy=copy)

    @overload
    def inverse_transform(self, X: np.ndarray, Y: None = None) -> np.ndarray | Universe | AtomGroup:
        ...

    @overload
    def inverse_transform(self, X: np.ndarray, Y: np.ndarray) -> tuple[np.ndarray | Universe | AtomGroup, np.ndarray]:
        ...

    def inverse_transform(self, X: np.ndarray, Y: Optional[np.ndarray] = None) -> \
            np.ndarray | Universe | AtomGroup | tuple[np.ndarray | Universe | AtomGroup, np.ndarray]:
        if Y is None:
            return self._from_crd(super().inverse_transform(X))
        else:
            X_new, Y_new = super().inverse_transform(X, Y)
            return self._from_crd(X_new), Y_new

    def inverse_predict(self,
                        Y: np.ndarray,
                        *,
                        ndim: Optional[int] = None,
                        copy=True):
        return self._to_crd(super().inverse_predict(Y, ndim=ndim, copy=copy))

    def predict(self,
                crd: np.ndarray | Universe | AtomGroup,
                *,
                ndim: Optional[int] = None,
                copy: bool = True):
        X = self._from_crd(crd)
        return super().predict(X, ndim=ndim, copy=copy)

    def score(self,
              crd: np.ndarray | Universe | AtomGroup,
              y: np.ndarray,
              sample_weight: Optional[np.ndarray] = None,
              *,
              ndim: Optional[int] = None,
              copy: bool = True) -> float:
        X = self._from_crd(crd)
        return super().score(X, y, sample_weight=sample_weight, ndim=ndim, copy=copy)


# A dirty hack to make the super functions be loaded only when type checking
if TYPE_CHECKING:
    class _Base_OPLS(OPLS, _MD_PLS_WRAPPER):
        ...
else:
    _Base_OPLS = _MD_PLS_WRAPPER


class _MD_OPLS_WRAPPER(_Base_OPLS):
    """
    A wrapper class to wrap OPLS object to accept MDAnalysis universes and atom groups,
    as well as unflattened MD trajectories. Should NOT be instanciated as such, but through the OPLS_MD class.
    The final class needs to first inherit this class, and only then the class which is wrapped.

    Only adds the wrapping for correct on top of _MD_PLS_WRAPPER
    """

    @overload
    def correct(self, crd: np.ndarray | Universe | AtomGroup, y: None = None, *,
                copy: bool = True,
                return_ortho: Literal[False] = False) -> np.ndarray:
        ...

    @overload
    def correct(self, crd: np.ndarray | Universe | AtomGroup, y: None = None, *,
                copy: bool = True,
                return_ortho: Literal[True] = True) -> tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def correct(self, crd: np.ndarray | Universe | AtomGroup, y: np.ndarray, *,
                copy: bool = True,
                return_ortho: Literal[False] = False) -> tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def correct(self, crd: np.ndarray | Universe | AtomGroup, y: np.ndarray, *,
                copy: bool = True,
                return_ortho: Literal[True] = True) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        ...

    @overload
    def correct(self, crd: np.ndarray | Universe | AtomGroup, y: Optional[np.ndarray] = None, *,
                copy: bool = True,
                return_ortho: bool = False) -> np.ndarray | tuple[np.ndarray, np.ndarray] |\
            tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        ...

    def correct(self,
                crd: np.ndarray | Universe | AtomGroup,
                y: Optional[np.ndarray] = None,
                *,
                copy: bool = True,
                return_ortho: bool = False) -> np.ndarray | tuple[np.ndarray, ...]:
        X = self._from_crd(crd)
        return super().correct(X, y, copy=copy, return_ortho=return_ortho)


class PLS_MD(_MD_PLS_WRAPPER, PLS):
    """
    Partial Least Squares (PLS), wrapped such that every method accepts an MDAnalysis
    Universe or AtomGroup, or an unflattened coordinate array of shape(n, n_atoms, ndim) instead of X-coordinates.
    In case of MDAnalysis, all frames in the trajectory of Universe (or AtomGroup.universe) will be read into memory.

    The output of the inverse transformations will also be reshaped back to the orginal dimensionality.

    Parameters
    ----------
    n_components : int, default=2
        Number of components to fit.
    scale : bool, default=True
        Whether to scale X and Y to unit variance
    center : bool, default=True
        Whether to centre X and Y to zero mean
    flip : bool, default=False
        Whether to flip the singular vectors for compatibility with different solvers.
        With flip=True and deflation_mode="regression" the PLS model will be the same
        (up to machine precision) as with sklearn.cross_decomposition.PLSRegression.
        With flip=False and univariate y, the x_scores_ are guaranteed to be positively
        correlated to y.
    max_iter : int, default=500
        Maximum number of NIPALS iterations.
    tol : float, default=1e-06
        Tolerance for stopping NIPALS iterations.
    copy : bool, default=True
        Whether to make copies of X and Y. False does not guarantee calculations are in place, but
        True does guarantee copying.
    dtype : numpy.dtype, default=float64
        numpy dtype to cast input to. If None calculations will be done with whatever dtype the input arrays have.
    deflation_mode : str, default=None
        Whether to calculate y deflation with x_scores ("regression") or y_scores ("canonical").
        The latter only is reliable with n_components<=Y.shape[1]. The first will make this the same as
        sklearn.cross_decomposition.PLSRegression and the latter the same as
        sklearn.cross_decomposition.PLSCanonical.
        In either case the first PLS component is the same.
    """
    pass


class OPLS_MD(_MD_OPLS_WRAPPER, OPLS):
    """
    Orthogonal Projection on Latent Structure (O-PLS), wrapped such that every method accepts an MDAnalysis
    Universe or AtomGroup, or an unflattened coordinate array of shape(n, n_atoms, ndim) instead of X-coordinates.
    In case of MDAnalysis, all frames in the trajectory of Universe (or AtomGroup.universe) will be read into memory.

    The output of the inverse transformations will also be reshaped back to the orginal dimensionality.

    If the input X is a numpy array with ndim=2, it is assumed to be already flattened.

    Parameters
    ----------
    n_components : int, default=1
        Number of components to fit.
    scale : bool, default=True
        Whether to scale X and Y to unit variance
    center : bool, default=True
        Whether to centre X and Y to zero mean
    flip : bool, default=False
        Whether to flip the singular vectors for compatibility with different solvers.
        With flip=True and deflation_mode="regression" the first PLS component will be
        exactly (up to machine precision) the same as with sklearn.cross_decomposition.PLSRegression.
        Does not affect the results in any other meaningful way.
    max_iter : int, default=500
        Maximum number of NIPALS iterations.
    tol : float, default=1e-06
        Tolerance for stopping NIPALS iterations.
    copy : bool, default=True
        Whether to make copies of X and Y. False does not guarantee everything is in place, but
        True does guarantee copying.
    algorithm : str, default="OPLS"
        The algorithm to use. Acceptable values are "OPLS" and "O2PLS".
        NOTE: "O2PLS" is not yet well tested, and OPLS is only tested with univariate y.
    deflation_mode : str, default=None
        Whether to calculate y deflation with x_scores ("regression") or y_scores ("canonical").
        The latter only is reliable with n_components<=Y.shape[1].
        With OPLS algorithm this only changes the y-loadings, y-rotations and the final regressor, so
        it does not affect the corrected coordinates.
        With O2PLS algorithm "canonical" should be used.
        If None, "regression" is used for OPLS and "canonical" with O2PLS.

    Attributes
    ----------
    predictive_scores: np.ndarray
        Predictive x-scores.
    predictive_loadings: np.ndarray
        Predictive x-loadings.
    orthogonal_scores: np.ndarray
        Orthogonal x-scores.
    orthogonal_loadings: np.ndarray
        Orthogonal x-loadings.
    """
    pass


class OPLS_PLS_MD(_MD_OPLS_WRAPPER, OPLS_PLS):
    """
    Orthogonal Projection on Latent Structure (O-PLS) wrapper for a Partial Least Squares (PLS)  model,
    wrapped such that every method accepts an MDAnalysis Universe or AtomGroup, or an unflattened coordinate
    array of shape(n, n_atoms, ndim) instead of X-coordinates. In case of MDAnalysis, all frames in the
    trajectory of Universe (or AtomGroup.universe) will be read into memory.

    The output of the inverse transformations will also be reshaped back to the orginal dimensionality.

    If the input X is a numpy array with ndim=2, it is assumed to be already flattened.

    Parameters
    ----------
    n_components : int, default=1
        Number of components to fit.
    pls_components : int, default=2
        Number of components to fit.
    scale : bool, default=True
        Whether to scale X and Y to unit variance
    center : bool, default=True
        Whether to centre X and Y to zero mean
    flip : bool, default=False
        Whether to flip the singular vectors for compatibility with different solvers.
        With flip=True and deflation_mode="regression" the first PLS component will be
        exactly (up to machine precision) the same as with sklearn.cross_decomposition.PLSRegression.
        Does not affect the results in any other meaningful way.
    max_iter : int, default=500
        Maximum number of NIPALS iterations.
    tol : float, default=1e-06
        Tolerance for stopping NIPALS iterations.
    copy : bool, default=True
        Whether to make copies of X and Y. False does not guarantee everything is in place, but
        True does guarantee copying.
    algorithm : str, default="OPLS"
        The algorithm to use. Acceptable values are "OPLS" and "O2PLS".
        NOTE: "O2PLS" is not yet well tested, and OPLS is only tested with univariate y.
    deflation_mode : str, default=None
        Whether to calculate y deflation with x_scores ("regression") or y_scores ("canonical").
        The latter only is reliable with n_components<=Y.shape[1].
        With OPLS algorithm this only changes the y-loadings, y-rotations and the final regressor, so
        it does not affect the corrected coordinates.
        With O2PLS algorithm "canonical" should be used.
        If None, "regression" is used for OPLS and "canonical" with O2PLS.

    Attributes
    ----------
    predictive_scores: np.ndarray
        Predictive x-scores.
    predictive_loadings: np.ndarray
        Predictive x-loadings.
    orthogonal_loadings: np.ndarray
        Orthogonal x-loadings.
    orthogonal_scores: np.ndarray
        Orthogonal x-scores.
    pls_: PLS
        The wrapped PLS model
    """

    @overload
    def transform_ortho(self, crd: np.ndarray | Universe | AtomGroup, Y: np.ndarray, *, copy: bool = True) -> tuple[np.ndarray, np.ndarray]:
        ...

    @overload
    def transform_ortho(self, crd: np.ndarray | Universe | AtomGroup, Y: None = None, *, copy: bool = True) -> np.ndarray:
        ...

    def transform_ortho(self,
                        crd: np.ndarray | Universe | AtomGroup,
                        Y: Optional[np.ndarray] = None, *,
                        copy: bool = True) -> tuple[np.ndarray, np.ndarray] | np.ndarray:
        X = self._from_crd(crd)
        return super().transform_ortho(X, Y, copy=copy)

    @overload
    def inverse_transform_ortho(self, X: np.ndarray, Y: None = None) -> np.ndarray:
        ...

    @overload
    def inverse_transform_ortho(self, X: np.ndarray, Y: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        ...

    def inverse_transform_ortho(self, X: np.ndarray, Y: Optional[np.ndarray] = None) -> tuple[np.ndarray, np.ndarray] | np.ndarray:
        if Y is None:
            return self._from_crd(super().inverse_transform_ortho(X))
        else:
            X_new, Y_new = super().inverse_transform_ortho(X, Y)
            return self._from_crd(X_new), Y_new
